<?php

namespace LakeDrops\Drupal8ScaffoldDeveloper;

use Composer\Composer;
use Composer\IO\IOInterface;
use Composer\Script\Event;
use LakeDrops\Docker4Drupal\Handler as D4D;
use Symfony\Component\Filesystem\Filesystem;

/**
 * Class Handler.
 *
 * @package LakeDrops\Drupal8ScaffoldDeveloper
 */
class Handler {

  /**
   * The composer object running this session.
   *
   * @var \Composer\Composer
   */
  protected $composer;

  /**
   * The input-output object for the composer session.
   *
   * @var \Composer\IO\IOInterface
   */
  protected $io;

  /**
   * The Drupal core package.
   *
   * @var \Composer\Package\PackageInterface
   */
  protected $drupalCorePackage;

  /**
   * Handler constructor.
   *
   * @param \Composer\Composer $composer
   *   The composer object.
   * @param \Composer\IO\IOInterface $io
   *   The input-output object.
   */
  public function __construct(Composer $composer, IOInterface $io) {
    $this->composer = $composer;
    $this->io = $io;
  }

  /**
   * Get the Drupal Core package.
   *
   * Look up the Drupal core package object, or return it from where we cached
   * it in the $drupalCorePackage field.
   *
   * @return \Composer\Package\PackageInterface
   *   The Drupal Core package.
   */
  protected function getDrupalCorePackage() {
    if (!isset($this->drupalCorePackage)) {
      $this->drupalCorePackage = $this->getPackage('drupal/core');
    }
    return $this->drupalCorePackage;
  }

  /**
   * Retrieve a package from the current composer process.
   *
   * @param string $name
   *   Name of the package to get from the current composer installation.
   *
   * @return \Composer\Package\PackageInterface
   *   The package.
   */
  protected function getPackage($name) {
    return $this->composer->getRepositoryManager()->getLocalRepository()->findPackage($name, '*');
  }

  /**
   * Post project create event to execute the scaffolding.
   *
   * @param \Composer\Script\Event $event
   *   The event that triggered the plugin.
   */
  public function setupLakeDropsProject(Event $event) {

    // We only do the fancy stuff for developers.
    if (!$event->isDevMode()) {
      return;
    }

    $options = $this->getOptions();
    $fs = new Filesystem();
    $installationManager = $this->composer->getInstallationManager();

    $drupalCorePackage = $this->getDrupalCorePackage();
    $corePath = $installationManager->getInstallPath($drupalCorePackage);

    // Directory where the root project is being created.
    $projectRoot = getcwd();
    // Directory where Drupal's index.php is located.
    $webRoot = dirname($corePath);
    // Directory where this plugin is being installed.
    $pluginRoot = $installationManager->getInstallPath($this->getPackage('lakedrops/d8-project-scaffold-developer'));

    // Create contrib/custom dirs for modules, profiles, themes and drush.
    foreach ([
      $webRoot . '/modules',
      $webRoot . '/profiles',
      $webRoot . '/themes',
      'drush',
    ] as $dir) {
      foreach (['contrib', 'custom'] as $subdir) {
        $path = $dir . '/' . $subdir;
        if (!$fs->exists($path)) {
          $fs->mkdir($path, 0775);
        }
        if ($options['chg-acl']) {
          $fs->chgrp($path, $options['webserver-groupname'], TRUE);
        }
      }
    }

    // Create and protect config, settings and files directories.
    foreach ([
      'config/default/sync',
      'settings/default',
      'files/default/files',
      'files/default/private',
    ] as $dir) {
      $path = $projectRoot . '/' . $dir;
      if (!$fs->exists($path)) {
        $fs->mkdir($path);
        $fs->copy($pluginRoot . '/templates/.htaccess', $path . '/.htaccess');
      }
      if ($options['chg-acl']) {
        $fs->chgrp($path, $options['webserver-groupname'], TRUE);
      }
    }

    // Link Drupal site's files and private directory.
    foreach (['files', 'private'] as $dir) {
      if (!$fs->exists($webRoot . '/sites/default/' . $dir)) {
        $rel = substr($fs->makePathRelative($projectRoot . '/files/default/' . $dir, $projectRoot . '/' . $webRoot . '/sites/default/' . $dir), 3, -1);
        $fs->symlink($rel, $webRoot . '/sites/default/' . $dir);
        $fs->chmod($webRoot . '/sites/default/' . $dir, 0775);
      }
    }

    // Link Drupal site's settings files.
    $options['hash'] = str_replace(['+', '/', '='], ['-', '_', ''], base64_encode(random_bytes(55)));
    $twig_loader = new \Twig_Loader_Array([]);
    $twig = new \Twig_Environment($twig_loader);
    foreach ([
      'settings.php',
      'services.yml',
      'development.services.yml',
      'settings.local.php',
    ] as $template) {
      $file = $projectRoot . '/settings/default/' . $template;
      $link = $webRoot . '/sites/default/' . $template;
      if (!$fs->exists($link)) {

        $twig_loader->setTemplate($template, file_get_contents($pluginRoot . '/templates/' . $template));
        $rendered = $twig->render($template, $options);
        file_put_contents($file, $rendered);

        $rel = substr($fs->makePathRelative($file, $projectRoot . '/' . $link), 3, -1);
        $fs->symlink($rel, $link);
      }
      $fs->chmod($file, 0664);
      if ($options['chg-acl']) {
        $fs->chgrp($file, $options['webserver-groupname']);
      }
    }

    // Copy Drush policy and readme.
    foreach (['policy.drush.inc', 'README.md'] as $template) {
      $fs->copy($pluginRoot . '/templates/drush/' . $template, 'drush/' . $template);
    }

    // Link Drupal site's config directory.
    if (!$fs->exists($webRoot . '/sites/default/files/config')) {
      $rel = substr($fs->makePathRelative($projectRoot . '/config/default', $projectRoot . '/files/default/files/config'), 3, -1);
      $fs->symlink($rel, $webRoot . '/sites/default/files/config');
    }

    // Tasks we only do the first time this is run.
    if (!$fs->exists($projectRoot . '/.git')) {

      // Remove template files.
      foreach (['.gitlab-ci.yml', 'README.md'] as $file) {
        $fs->remove($file);
      }

      // Initialize local git working copy.
      try {
        $this->git('init');
        $this->git('add *');
        $this->git('commit -m "Initial commit"');
      }
      catch (\Exception $ex) {
        // We're ignoring this for now.
      }
    }

    // If available, also configure Docker4Drupal.
    if ($this->getPackage('lakedrops/docker4drupal')) {
      $handler = new D4D($event->getComposer(), $event->getIO());
      $handler->configureProject($event);
    }
  }

  /**
   * Retrieve configuration for this package.
   *
   * @return array
   *   The settings from the extra configuration.
   */
  protected function getOptions() {
    $extra = $this->composer->getPackage()->getExtra() + ['d8-project-scaffold' => []];
    $options = $extra['d8-project-scaffold'] + [
      'chg-acl' => FALSE,
      'webserver-username' => 'www-data',
      'webserver-groupname' => 'www-data',
      'filter_protocols' => [
        'http',
        'https',
        'ftp',
        'news',
        'nntp',
        'tel',
        'telnet',
        'mailto',
        'irc',
        'ssh',
        'sftp',
        'webcal',
        'rtsp',
      ],
      'cors_config' => [
        'enabled' => 'false',
        'allowedOrigins' => '*',
      ],
    ];
    return $options;
  }

  /**
   * Wrapper for git command in the root directory.
   *
   * @param string $command
   *   Git command name, arguments and/or options.
   *
   * @throws \Exception
   */
  protected function git($command) {
    passthru(escapeshellcmd('git -c "user.email=d8-project@lakedrops.com" ' . $command), $exit_code);
    if ($exit_code !== 0) {
      throw new \Exception('Git returned a non-zero exit code');
    }
  }

}
